# coding=utf-8
"""Modules taken from OpenStudio-standards.

https://github.com/NREL/openstudio-standards/blob/master/
lib/openstudio-standards/prototypes/common/objects/Prototype.CoilCoolingWater.rb
"""
from __future__ import division

from honeybee_openstudio.openstudio import openstudio, openstudio_model

from .utilities import create_curve_biquadratic, create_curve_quadratic, \
    convert_curve_biquadratic
from .schedule import model_add_schedule


def create_coil_cooling_water(
        model, chilled_water_loop, air_loop_node=None, name='Clg Coil', schedule=None,
        design_inlet_water_temperature=None, design_inlet_air_temperature=None,
        design_outlet_air_temperature=None):
    clg_coil = openstudio_model.CoilCoolingWater(model)

    # add to chilled water loop
    chilled_water_loop.addDemandBranchForComponent(clg_coil)

    # add to air loop if specified
    if air_loop_node is not None:
        clg_coil.addToNode(air_loop_node)

    # set coil name
    name = 'Clg Coil' if name is None else name
    clg_coil.setName(name)

    # set coil availability schedule
    if schedule is None:  # default always on
        coil_availability_schedule = model.alwaysOnDiscreteSchedule()
    elif isinstance(schedule, str):
        if schedule == 'alwaysOffDiscreteSchedule':
            coil_availability_schedule = model.alwaysOffDiscreteSchedule()
        elif schedule == 'alwaysOnDiscreteSchedule':
            coil_availability_schedule = model.alwaysOnDiscreteSchedule()
        else:
            coil_availability_schedule = model_add_schedule(model, schedule)
    else:  # assume that it is an actual schedule object
        coil_availability_schedule = schedule
    clg_coil.setAvailabilitySchedule(coil_availability_schedule)

    # rated temperatures
    if design_inlet_water_temperature is None:
        clg_coil.autosizeDesignInletWaterTemperature()
    else:
        clg_coil.setDesignInletWaterTemperature(design_inlet_water_temperature)
    if design_inlet_air_temperature is not None:
        clg_coil.setDesignInletAirTemperature(design_inlet_air_temperature)
    if design_outlet_air_temperature is not None:
        clg_coil.setDesignOutletAirTemperature(design_outlet_air_temperature)

    # defaults
    clg_coil.setHeatExchangerConfiguration('CrossFlow')

    # coil controller properties
    # @note These inputs will get overwritten if addToNode or addDemandBranchForComponent
    # is called on the htg_coil object after this
    clg_coil_controller = clg_coil.controllerWaterCoil().get()
    clg_coil_controller.setName('{} Controller'.format(clg_coil.nameString()))
    clg_coil_controller.setAction('Reverse')
    clg_coil_controller.setMinimumActuatedFlow(0.0)

    return clg_coil


def create_coil_cooling_dx_single_speed(
        model, air_loop_node=None, name='1spd DX Clg Coil', schedule=None,
        type=None, cop=None):
    """Prototype CoilCoolingDXSingleSpeed object.

    Enters in default curves for coil by type of coil

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        air_loop_node: [<OpenStudio::Model::Node>] the coil will be placed on
            this node of the air loop.
        name: [String] the name of the system, or nil in which case it will be defaulted.
        schedule: [String] name of the availability schedule, or
            [<OpenStudio::Model::Schedule>] Schedule object, or nil in which case
            default to always on.
        type: [String] the type of single speed DX coil to reference the
            correct curve set.
        cop: [Double] rated cooling coefficient of performance.
    """
    clg_coil = openstudio_model.CoilCoolingDXSingleSpeed(model)

    # add to air loop if specified
    if air_loop_node is not None:
        clg_coil.addToNode(air_loop_node)

    # set coil name
    clg_coil.setName(name)

    # set coil availability schedule
    if schedule is None:  # default always on
        coil_availability_schedule = model.alwaysOnDiscreteSchedule()
    elif isinstance(schedule, str):
        if schedule == 'alwaysOffDiscreteSchedule':
            coil_availability_schedule = model.alwaysOffDiscreteSchedule()
        elif schedule == 'alwaysOnDiscreteSchedule':
            coil_availability_schedule = model.alwaysOnDiscreteSchedule()
        else:
            coil_availability_schedule = model_add_schedule(model, schedule)
    else:  # assume that it is an actual schedule object
        coil_availability_schedule = schedule
    clg_coil.setAvailabilitySchedule(coil_availability_schedule)

    # set coil cop
    if cop is not None:
        clg_coil.setRatedCOP(cop)

    clg_cap_f_of_temp = None
    clg_cap_f_of_flow = None
    clg_energy_input_ratio_f_of_temp = None
    clg_energy_input_ratio_f_of_flow = None
    clg_part_load_ratio = None

    # curve sets
    if type == 'OS default':
        pass  # use OS defaults

    elif type == 'Heat Pump':
        # "PSZ-AC_Unitary_PackagecoolCapFT"
        clg_cap_f_of_temp = openstudio_model.CurveBiquadratic(model)
        clg_cap_f_of_temp.setCoefficient1Constant(0.766956)
        clg_cap_f_of_temp.setCoefficient2x(0.0107756)
        clg_cap_f_of_temp.setCoefficient3xPOW2(-0.0000414703)
        clg_cap_f_of_temp.setCoefficient4y(0.00134961)
        clg_cap_f_of_temp.setCoefficient5yPOW2(-0.000261144)
        clg_cap_f_of_temp.setCoefficient6xTIMESY(0.000457488)
        clg_cap_f_of_temp.setMinimumValueofx(12.78)
        clg_cap_f_of_temp.setMaximumValueofx(23.89)
        clg_cap_f_of_temp.setMinimumValueofy(21.1)
        clg_cap_f_of_temp.setMaximumValueofy(46.1)

        clg_cap_f_of_flow = openstudio_model.CurveQuadratic(model)
        clg_cap_f_of_flow.setCoefficient1Constant(0.8)
        clg_cap_f_of_flow.setCoefficient2x(0.2)
        clg_cap_f_of_flow.setCoefficient3xPOW2(0.0)
        clg_cap_f_of_flow.setMinimumValueofx(0.5)
        clg_cap_f_of_flow.setMaximumValueofx(1.5)

        clg_energy_input_ratio_f_of_temp = openstudio_model.CurveBiquadratic(model)
        clg_energy_input_ratio_f_of_temp.setCoefficient1Constant(0.297145)
        clg_energy_input_ratio_f_of_temp.setCoefficient2x(0.0430933)
        clg_energy_input_ratio_f_of_temp.setCoefficient3xPOW2(-0.000748766)
        clg_energy_input_ratio_f_of_temp.setCoefficient4y(0.00597727)
        clg_energy_input_ratio_f_of_temp.setCoefficient5yPOW2(0.000482112)
        clg_energy_input_ratio_f_of_temp.setCoefficient6xTIMESY(-0.000956448)
        clg_energy_input_ratio_f_of_temp.setMinimumValueofx(12.78)
        clg_energy_input_ratio_f_of_temp.setMaximumValueofx(23.89)
        clg_energy_input_ratio_f_of_temp.setMinimumValueofy(21.1)
        clg_energy_input_ratio_f_of_temp.setMaximumValueofy(46.1)

        clg_energy_input_ratio_f_of_flow = openstudio_model.CurveQuadratic(model)
        clg_energy_input_ratio_f_of_flow.setCoefficient1Constant(1.156)
        clg_energy_input_ratio_f_of_flow.setCoefficient2x(-0.1816)
        clg_energy_input_ratio_f_of_flow.setCoefficient3xPOW2(0.0256)
        clg_energy_input_ratio_f_of_flow.setMinimumValueofx(0.5)
        clg_energy_input_ratio_f_of_flow.setMaximumValueofx(1.5)

        clg_part_load_ratio = openstudio_model.CurveQuadratic(model)
        clg_part_load_ratio.setCoefficient1Constant(0.85)
        clg_part_load_ratio.setCoefficient2x(0.15)
        clg_part_load_ratio.setCoefficient3xPOW2(0.0)
        clg_part_load_ratio.setMinimumValueofx(0.0)
        clg_part_load_ratio.setMaximumValueofx(1.0)

    elif type == 'PSZ-AC':
        # Defaults to "DOE Ref DX Clg Coil Cool-Cap-fT"
        clg_cap_f_of_temp = openstudio_model.CurveBiquadratic(model)
        clg_cap_f_of_temp.setCoefficient1Constant(0.9712123)
        clg_cap_f_of_temp.setCoefficient2x(-0.015275502)
        clg_cap_f_of_temp.setCoefficient3xPOW2(0.0014434524)
        clg_cap_f_of_temp.setCoefficient4y(-0.00039321)
        clg_cap_f_of_temp.setCoefficient5yPOW2(-0.0000068364)
        clg_cap_f_of_temp.setCoefficient6xTIMESY(-0.0002905956)
        clg_cap_f_of_temp.setMinimumValueofx(-100.0)
        clg_cap_f_of_temp.setMaximumValueofx(100.0)
        clg_cap_f_of_temp.setMinimumValueofy(-100.0)
        clg_cap_f_of_temp.setMaximumValueofy(100.0)

        clg_cap_f_of_flow = openstudio_model.CurveQuadratic(model)
        clg_cap_f_of_flow.setCoefficient1Constant(1.0)
        clg_cap_f_of_flow.setCoefficient2x(0.0)
        clg_cap_f_of_flow.setCoefficient3xPOW2(0.0)
        clg_cap_f_of_flow.setMinimumValueofx(-100.0)
        clg_cap_f_of_flow.setMaximumValueofx(100.0)

        # "DOE Ref DX Clg Coil Cool-EIR-fT",
        clg_energy_input_ratio_f_of_temp = openstudio_model.CurveBiquadratic(model)
        clg_energy_input_ratio_f_of_temp.setCoefficient1Constant(0.28687133)
        clg_energy_input_ratio_f_of_temp.setCoefficient2x(0.023902164)
        clg_energy_input_ratio_f_of_temp.setCoefficient3xPOW2(-0.000810648)
        clg_energy_input_ratio_f_of_temp.setCoefficient4y(0.013458546)
        clg_energy_input_ratio_f_of_temp.setCoefficient5yPOW2(0.0003389364)
        clg_energy_input_ratio_f_of_temp.setCoefficient6xTIMESY(-0.0004870044)
        clg_energy_input_ratio_f_of_temp.setMinimumValueofx(-100.0)
        clg_energy_input_ratio_f_of_temp.setMaximumValueofx(100.0)
        clg_energy_input_ratio_f_of_temp.setMinimumValueofy(-100.0)
        clg_energy_input_ratio_f_of_temp.setMaximumValueofy(100.0)

        clg_energy_input_ratio_f_of_flow = openstudio_model.CurveQuadratic(model)
        clg_energy_input_ratio_f_of_flow.setCoefficient1Constant(1.0)
        clg_energy_input_ratio_f_of_flow.setCoefficient2x(0.0)
        clg_energy_input_ratio_f_of_flow.setCoefficient3xPOW2(0.0)
        clg_energy_input_ratio_f_of_flow.setMinimumValueofx(-100.0)
        clg_energy_input_ratio_f_of_flow.setMaximumValueofx(100.0)

        # "DOE Ref DX Clg Coil Cool-PLF-fPLR"
        clg_part_load_ratio = openstudio_model.CurveQuadratic(model)
        clg_part_load_ratio.setCoefficient1Constant(0.90949556)
        clg_part_load_ratio.setCoefficient2x(0.09864773)
        clg_part_load_ratio.setCoefficient3xPOW2(-0.00819488)
        clg_part_load_ratio.setMinimumValueofx(0.0)
        clg_part_load_ratio.setMaximumValueofx(1.0)
        clg_part_load_ratio.setMinimumCurveOutput(0.7)
        clg_part_load_ratio.setMaximumCurveOutput(1.0)

    elif type == 'Window AC':
        # Performance curves
        # From Frigidaire 10.7 EER unit in Winkler et. al. Lab Testing of Window ACs (2013)
        # @note These coefficients are in SI UNITS
        cool_cap_ft_coeffs_si = [0.6405, 0.01568, 0.0004531,
                                 0.001615, -0.0001825, 0.00006614]
        cool_eir_ft_coeffs_si = [2.287, -0.1732, 0.004745, 0.01662, 0.000484, -0.001306]
        cool_cap_fflow_coeffs = [0.887, 0.1128, 0]
        cool_eir_fflow_coeffs = [1.763, -0.6081, 0]
        cool_plf_fplr_coeffs = [0.78, 0.22, 0]

        # Make the curves
        clg_cap_f_of_temp = create_curve_biquadratic(
            model, cool_cap_ft_coeffs_si, 'RoomAC-Cap-fT', 0, 100, 0, 100, None, None)
        clg_cap_f_of_flow = create_curve_quadratic(
            model, cool_cap_fflow_coeffs, 'RoomAC-Cap-fFF', 0, 2, 0, 2,
            is_dimensionless=True)
        clg_energy_input_ratio_f_of_temp = create_curve_biquadratic(
            model, cool_eir_ft_coeffs_si, 'RoomAC-EIR-fT', 0, 100, 0, 100, None, None)
        clg_energy_input_ratio_f_of_flow = create_curve_quadratic(
            model, cool_eir_fflow_coeffs, 'RoomAC-EIR-fFF', 0, 2, 0, 2,
            is_dimensionless=True)
        clg_part_load_ratio = create_curve_quadratic(
            model, cool_plf_fplr_coeffs, 'RoomAC-PLF-fPLR', 0, 1, 0, 1,
            is_dimensionless=True)

    elif type == 'Residential Central AC':
        # Performance curves
        # These coefficients are in IP UNITS
        cool_cap_ft_coeffs_ip = [3.670270705, -0.098652414, 0.000955906,
                                 0.006552414, -0.0000156, -0.000131877]
        cool_eir_ft_coeffs_ip = [-3.302695861, 0.137871531, -0.001056996,
                                 -0.012573945, 0.000214638, -0.000145054]
        cool_cap_fflow_coeffs = [0.718605468, 0.410099989, -0.128705457]
        cool_eir_fflow_coeffs = [1.32299905, -0.477711207, 0.154712157]
        cool_plf_fplr_coeffs = [0.8, 0.2, 0]

        # Convert coefficients from IP to SI
        cool_cap_ft_coeffs_si = convert_curve_biquadratic(cool_cap_ft_coeffs_ip)
        cool_eir_ft_coeffs_si = convert_curve_biquadratic(cool_eir_ft_coeffs_ip)

        # Make the curves
        clg_cap_f_of_temp = create_curve_biquadratic(
            model, cool_cap_ft_coeffs_si, 'AC-Cap-fT', 0, 100, 0, 100, None, None)
        clg_cap_f_of_flow = create_curve_quadratic(
            model, cool_cap_fflow_coeffs, 'AC-Cap-fFF', 0, 2, 0, 2, is_dimensionless=True)
        clg_energy_input_ratio_f_of_temp = create_curve_biquadratic(
            model, cool_eir_ft_coeffs_si, 'AC-EIR-fT', 0, 100, 0, 100, None, None)
        clg_energy_input_ratio_f_of_flow = create_curve_quadratic(
            model, cool_eir_fflow_coeffs, 'AC-EIR-fFF', 0, 2, 0, 2, is_dimensionless=True)
        clg_part_load_ratio = create_curve_quadratic(
            model, cool_plf_fplr_coeffs, 'AC-PLF-fPLR', 0, 1, 0, 1, is_dimensionless=True)

    elif type == 'Residential Central ASHP':
        # Performance curves
        # These coefficients are in IP UNITS
        cool_cap_ft_coeffs_ip = [3.68637657, -0.098352478, 0.000956357,
                                 0.005838141, -0.0000127, -0.000131702]
        cool_eir_ft_coeffs_ip = [-3.437356399, 0.136656369, -0.001049231,
                                 -0.0079378, 0.000185435, -0.0001441]
        cool_cap_fflow_coeffs = [0.718664047, 0.41797409, -0.136638137]
        cool_eir_fflow_coeffs = [1.143487507, -0.13943972, -0.004047787]
        cool_plf_fplr_coeffs = [0.8, 0.2, 0]

        # Convert coefficients from IP to SI
        cool_cap_ft_coeffs_si = convert_curve_biquadratic(cool_cap_ft_coeffs_ip)
        cool_eir_ft_coeffs_si = convert_curve_biquadratic(cool_eir_ft_coeffs_ip)

        # Make the curves
        clg_cap_f_of_temp = create_curve_biquadratic(
            model, cool_cap_ft_coeffs_si, 'Cool-Cap-fT', 0, 100, 0, 100, None, None)
        clg_cap_f_of_flow = create_curve_quadratic(
            model, cool_cap_fflow_coeffs, 'Cool-Cap-fFF', 0, 2, 0, 2,
            is_dimensionless=True)
        clg_energy_input_ratio_f_of_temp = create_curve_biquadratic(
            model, cool_eir_ft_coeffs_si, 'Cool-EIR-fT', 0, 100, 0, 100, None, None)
        clg_energy_input_ratio_f_of_flow = create_curve_quadratic(
            model, cool_eir_fflow_coeffs, 'Cool-EIR-fFF', 0, 2, 0, 2,
            is_dimensionless=True)
        clg_part_load_ratio = create_curve_quadratic(
            model, cool_plf_fplr_coeffs, 'Cool-PLF-fPLR', 0, 1, 0, 1,
            is_dimensionless=True)

    else:  # default curve set, type == 'Split AC' || 'PTAC'
        clg_cap_f_of_temp = openstudio_model.CurveBiquadratic(model)
        clg_cap_f_of_temp.setCoefficient1Constant(0.942587793)
        clg_cap_f_of_temp.setCoefficient2x(0.009543347)
        clg_cap_f_of_temp.setCoefficient3xPOW2(0.00068377)
        clg_cap_f_of_temp.setCoefficient4y(-0.011042676)
        clg_cap_f_of_temp.setCoefficient5yPOW2(0.000005249)
        clg_cap_f_of_temp.setCoefficient6xTIMESY(-0.00000972)
        clg_cap_f_of_temp.setMinimumValueofx(12.77778)
        clg_cap_f_of_temp.setMaximumValueofx(23.88889)
        clg_cap_f_of_temp.setMinimumValueofy(23.88889)
        clg_cap_f_of_temp.setMaximumValueofy(46.11111)

        clg_cap_f_of_flow = openstudio_model.CurveQuadratic(model)
        clg_cap_f_of_flow.setCoefficient1Constant(0.8)
        clg_cap_f_of_flow.setCoefficient2x(0.2)
        clg_cap_f_of_flow.setCoefficient3xPOW2(0)
        clg_cap_f_of_flow.setMinimumValueofx(0.5)
        clg_cap_f_of_flow.setMaximumValueofx(1.5)

        clg_energy_input_ratio_f_of_temp = openstudio_model.CurveBiquadratic(model)
        clg_energy_input_ratio_f_of_temp.setCoefficient1Constant(0.342414409)
        clg_energy_input_ratio_f_of_temp.setCoefficient2x(0.034885008)
        clg_energy_input_ratio_f_of_temp.setCoefficient3xPOW2(-0.0006237)
        clg_energy_input_ratio_f_of_temp.setCoefficient4y(0.004977216)
        clg_energy_input_ratio_f_of_temp.setCoefficient5yPOW2(0.000437951)
        clg_energy_input_ratio_f_of_temp.setCoefficient6xTIMESY(-0.000728028)
        clg_energy_input_ratio_f_of_temp.setMinimumValueofx(12.77778)
        clg_energy_input_ratio_f_of_temp.setMaximumValueofx(23.88889)
        clg_energy_input_ratio_f_of_temp.setMinimumValueofy(23.88889)
        clg_energy_input_ratio_f_of_temp.setMaximumValueofy(46.11111)

        clg_energy_input_ratio_f_of_flow = openstudio_model.CurveQuadratic(model)
        clg_energy_input_ratio_f_of_flow.setCoefficient1Constant(1.1552)
        clg_energy_input_ratio_f_of_flow.setCoefficient2x(-0.1808)
        clg_energy_input_ratio_f_of_flow.setCoefficient3xPOW2(0.0256)
        clg_energy_input_ratio_f_of_flow.setMinimumValueofx(0.5)
        clg_energy_input_ratio_f_of_flow.setMaximumValueofx(1.5)

        clg_part_load_ratio = openstudio_model.CurveQuadratic(model)
        clg_part_load_ratio.setCoefficient1Constant(0.85)
        clg_part_load_ratio.setCoefficient2x(0.15)
        clg_part_load_ratio.setCoefficient3xPOW2(0.0)
        clg_part_load_ratio.setMinimumValueofx(0.0)
        clg_part_load_ratio.setMaximumValueofx(1.0)
        clg_part_load_ratio.setMinimumCurveOutput(0.7)
        clg_part_load_ratio.setMaximumCurveOutput(1.0)

    if clg_cap_f_of_temp is not None:
        clg_coil.setTotalCoolingCapacityFunctionOfTemperatureCurve(clg_cap_f_of_temp)
    if clg_cap_f_of_flow is not None:
        clg_coil.setTotalCoolingCapacityFunctionOfFlowFractionCurve(clg_cap_f_of_flow)
    if clg_energy_input_ratio_f_of_temp is not None:
        clg_coil.setEnergyInputRatioFunctionOfTemperatureCurve(
            clg_energy_input_ratio_f_of_temp)
    if clg_energy_input_ratio_f_of_flow is not None:
        clg_coil.setEnergyInputRatioFunctionOfFlowFractionCurve(
            clg_energy_input_ratio_f_of_flow)
    if clg_part_load_ratio is not None:
        clg_coil.setPartLoadFractionCorrelationCurve(clg_part_load_ratio)

    return clg_coil


def create_coil_cooling_dx_two_speed(
        model, air_loop_node=None, name='2spd DX Clg Coil', schedule=None, type=None):
    """Prototype CoilCoolingDXTwoSpeed object.

    Enters in default curves for coil by type of coil

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        air_loop_node: [<OpenStudio::Model::Node>] the coil will be placed on
            this node of the air loop.
        name: [String] the name of the system, or nil in which case it will be defaulted.
        schedule: [String] name of the availability schedule, or
            [<OpenStudio::Model::Schedule>] Schedule object, or nil in which case
            default to always on.
        type: [String] the type of two speed DX coil to reference the correct curve set.
    """
    clg_coil = openstudio_model.CoilCoolingDXTwoSpeed(model)

    # add to air loop if specified
    if air_loop_node is not None:
        clg_coil.addToNode(air_loop_node)

    # set coil name
    name = '2spd DX Clg Coil' if name is None else name
    clg_coil.setName(name)

    # set coil availability schedule
    if schedule is None:  # default always on
        coil_availability_schedule = model.alwaysOnDiscreteSchedule()
    elif isinstance(schedule, str):
        if schedule == 'alwaysOffDiscreteSchedule':
            coil_availability_schedule = model.alwaysOffDiscreteSchedule()
        elif schedule == 'alwaysOnDiscreteSchedule':
            coil_availability_schedule = model.alwaysOnDiscreteSchedule()
        else:
            coil_availability_schedule = model_add_schedule(model, schedule)
    else:  # assume that it is an actual schedule object
        coil_availability_schedule = schedule
    clg_coil.setAvailabilitySchedule(coil_availability_schedule)

    clg_cap_f_of_temp = None
    clg_cap_f_of_flow = None
    clg_energy_input_ratio_f_of_temp = None
    clg_energy_input_ratio_f_of_flow = None
    clg_part_load_ratio = None
    clg_cap_f_of_temp_low_spd = None
    clg_energy_input_ratio_f_of_temp_low_spd = None

    # curve sets
    if type == 'OS default':
        pass  # use OS defaults
    elif type == 'Residential Minisplit HP':
        # Performance curves
        # These coefficients are in SI units
        cool_cap_ft_coeffs_si = [0.7531983499655835, 0.003618193903031667, 0.0,
                                 0.006574385031351544, -6.87181191015432e-05, 0.0]
        cool_eir_ft_coeffs_si = [-0.06376924779982301, -0.0013360593470367282,
                                 1.413060577993827e-05, 0.019433076486584752,
                                 -4.91395947154321e-05, -4.909341249475308e-05]
        cool_cap_fflow_coeffs = [1, 0, 0]
        cool_eir_fflow_coeffs = [1, 0, 0]
        cool_plf_fplr_coeffs = [0.89, 0.11, 0]

        # Make the curves
        clg_cap_f_of_temp = create_curve_biquadratic(
            model, cool_cap_ft_coeffs_si, 'Cool-Cap-fT', 0, 100, 0, 100, None, None)
        clg_cap_f_of_flow = create_curve_quadratic(
            model, cool_cap_fflow_coeffs, 'Cool-Cap-fFF', 0, 2, 0, 2, is_dimensionless=True)
        clg_energy_input_ratio_f_of_temp = create_curve_biquadratic(
            model, cool_eir_ft_coeffs_si, 'Cool-EIR-fT', 0, 100, 0, 100, None, None)
        clg_energy_input_ratio_f_of_flow = create_curve_quadratic(
            model, cool_eir_fflow_coeffs, 'Cool-EIR-fFF', 0, 2, 0, 2, is_dimensionless=True)
        clg_part_load_ratio = create_curve_quadratic(
            model, cool_plf_fplr_coeffs, 'Cool-PLF-fPLR', 0, 1, 0, 1, is_dimensionless=True)
        clg_cap_f_of_temp_low_spd = create_curve_biquadratic(
            model, cool_cap_ft_coeffs_si, 'Cool-Cap-fT', 0, 100, 0, 100, None, None)
        clg_energy_input_ratio_f_of_temp_low_spd = create_curve_biquadratic(
            model, cool_eir_ft_coeffs_si, 'Cool-EIR-fT', 0, 100, 0, 100, None, None)
        clg_coil.setRatedLowSpeedSensibleHeatRatio(0.73)
        clg_coil.setCondenserType('AirCooled')
    else:  # default curve set, type == 'PSZ-AC' || 'Split AC' || 'PTAC'
        clg_cap_f_of_temp = openstudio_model.CurveBiquadratic(model)
        clg_cap_f_of_temp.setCoefficient1Constant(0.42415)
        clg_cap_f_of_temp.setCoefficient2x(0.04426)
        clg_cap_f_of_temp.setCoefficient3xPOW2(-0.00042)
        clg_cap_f_of_temp.setCoefficient4y(0.00333)
        clg_cap_f_of_temp.setCoefficient5yPOW2(-0.00008)
        clg_cap_f_of_temp.setCoefficient6xTIMESY(-0.00021)
        clg_cap_f_of_temp.setMinimumValueofx(17.0)
        clg_cap_f_of_temp.setMaximumValueofx(22.0)
        clg_cap_f_of_temp.setMinimumValueofy(13.0)
        clg_cap_f_of_temp.setMaximumValueofy(46.0)

        clg_cap_f_of_flow = openstudio_model.CurveQuadratic(model)
        clg_cap_f_of_flow.setCoefficient1Constant(0.77136)
        clg_cap_f_of_flow.setCoefficient2x(0.34053)
        clg_cap_f_of_flow.setCoefficient3xPOW2(-0.11088)
        clg_cap_f_of_flow.setMinimumValueofx(0.75918)
        clg_cap_f_of_flow.setMaximumValueofx(1.13877)

        clg_energy_input_ratio_f_of_temp = openstudio_model.CurveBiquadratic(model)
        clg_energy_input_ratio_f_of_temp.setCoefficient1Constant(1.23649)
        clg_energy_input_ratio_f_of_temp.setCoefficient2x(-0.02431)
        clg_energy_input_ratio_f_of_temp.setCoefficient3xPOW2(0.00057)
        clg_energy_input_ratio_f_of_temp.setCoefficient4y(-0.01434)
        clg_energy_input_ratio_f_of_temp.setCoefficient5yPOW2(0.00063)
        clg_energy_input_ratio_f_of_temp.setCoefficient6xTIMESY(-0.00038)
        clg_energy_input_ratio_f_of_temp.setMinimumValueofx(17.0)
        clg_energy_input_ratio_f_of_temp.setMaximumValueofx(22.0)
        clg_energy_input_ratio_f_of_temp.setMinimumValueofy(13.0)
        clg_energy_input_ratio_f_of_temp.setMaximumValueofy(46.0)

        clg_energy_input_ratio_f_of_flow = openstudio_model.CurveQuadratic(model)
        clg_energy_input_ratio_f_of_flow.setCoefficient1Constant(1.20550)
        clg_energy_input_ratio_f_of_flow.setCoefficient2x(-0.32953)
        clg_energy_input_ratio_f_of_flow.setCoefficient3xPOW2(0.12308)
        clg_energy_input_ratio_f_of_flow.setMinimumValueofx(0.75918)
        clg_energy_input_ratio_f_of_flow.setMaximumValueofx(1.13877)

        clg_part_load_ratio = openstudio_model.CurveQuadratic(model)
        clg_part_load_ratio.setCoefficient1Constant(0.77100)
        clg_part_load_ratio.setCoefficient2x(0.22900)
        clg_part_load_ratio.setCoefficient3xPOW2(0.0)
        clg_part_load_ratio.setMinimumValueofx(0.0)
        clg_part_load_ratio.setMaximumValueofx(1.0)

        clg_cap_f_of_temp_low_spd = openstudio_model.CurveBiquadratic(model)
        clg_cap_f_of_temp_low_spd.setCoefficient1Constant(0.42415)
        clg_cap_f_of_temp_low_spd.setCoefficient2x(0.04426)
        clg_cap_f_of_temp_low_spd.setCoefficient3xPOW2(-0.00042)
        clg_cap_f_of_temp_low_spd.setCoefficient4y(0.00333)
        clg_cap_f_of_temp_low_spd.setCoefficient5yPOW2(-0.00008)
        clg_cap_f_of_temp_low_spd.setCoefficient6xTIMESY(-0.00021)
        clg_cap_f_of_temp_low_spd.setMinimumValueofx(17.0)
        clg_cap_f_of_temp_low_spd.setMaximumValueofx(22.0)
        clg_cap_f_of_temp_low_spd.setMinimumValueofy(13.0)
        clg_cap_f_of_temp_low_spd.setMaximumValueofy(46.0)

        clg_energy_input_ratio_f_of_temp_low_spd = openstudio_model.CurveBiquadratic(model)
        clg_energy_input_ratio_f_of_temp_low_spd.setCoefficient1Constant(1.23649)
        clg_energy_input_ratio_f_of_temp_low_spd.setCoefficient2x(-0.02431)
        clg_energy_input_ratio_f_of_temp_low_spd.setCoefficient3xPOW2(0.00057)
        clg_energy_input_ratio_f_of_temp_low_spd.setCoefficient4y(-0.01434)
        clg_energy_input_ratio_f_of_temp_low_spd.setCoefficient5yPOW2(0.00063)
        clg_energy_input_ratio_f_of_temp_low_spd.setCoefficient6xTIMESY(-0.00038)
        clg_energy_input_ratio_f_of_temp_low_spd.setMinimumValueofx(17.0)
        clg_energy_input_ratio_f_of_temp_low_spd.setMaximumValueofx(22.0)
        clg_energy_input_ratio_f_of_temp_low_spd.setMinimumValueofy(13.0)
        clg_energy_input_ratio_f_of_temp_low_spd.setMaximumValueofy(46.0)

        clg_coil.setRatedLowSpeedSensibleHeatRatio(0.69)
        clg_coil.setBasinHeaterCapacity(10)
        clg_coil.setBasinHeaterSetpointTemperature(2.0)

    if clg_cap_f_of_temp is not None:
        clg_coil.setTotalCoolingCapacityFunctionOfTemperatureCurve(clg_cap_f_of_temp)
    if clg_cap_f_of_flow is not None:
        clg_coil.setTotalCoolingCapacityFunctionOfFlowFractionCurve(clg_cap_f_of_flow)
    if clg_energy_input_ratio_f_of_temp is not None:
        clg_coil.setEnergyInputRatioFunctionOfTemperatureCurve(
            clg_energy_input_ratio_f_of_temp)
    if clg_energy_input_ratio_f_of_flow is not None:
        clg_coil.setEnergyInputRatioFunctionOfFlowFractionCurve(
            clg_energy_input_ratio_f_of_flow)
    if clg_part_load_ratio is not None:
        clg_coil.setPartLoadFractionCorrelationCurve(clg_part_load_ratio)
    if clg_cap_f_of_temp_low_spd is not None:
        clg_coil.setLowSpeedTotalCoolingCapacityFunctionOfTemperatureCurve(
            clg_cap_f_of_temp_low_spd)
    if clg_energy_input_ratio_f_of_temp_low_spd is not None:
        clg_coil.setLowSpeedEnergyInputRatioFunctionOfTemperatureCurve(
            clg_energy_input_ratio_f_of_temp_low_spd)

    return clg_coil


def create_coil_cooling_water_to_air_heat_pump_equation_fit(
        model, plant_loop, air_loop_node=None, name='Water-to-Air HP Clg Coil',
        type=None, cop=3.4):
    """Prototype CoilCoolingWaterToAirHeatPumpEquationFit object.

    Enters in default curves for coil by type of coil.

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        plant_loop: [<OpenStudio::Model::PlantLoop>] the coil will be placed on
            the demand side of this plant loop.
        air_loop_node: [<OpenStudio::Model::Node>] the coil will be placed on
            this node of the air loop.
        name: [String] the name of the system, or nil in which case it will
            be defaulted.
        type: [String] the type of coil to reference the correct curve set.
        cop: [Double] rated cooling coefficient of performance.
    """
    clg_coil = openstudio_model.CoilCoolingWaterToAirHeatPumpEquationFit(model)

    # add to air loop if specified
    if air_loop_node is not None:
        clg_coil.addToNode(air_loop_node)

    # set coil name
    clg_coil.setName(name)

    # add to plant loop
    if plant_loop is None:
        raise ValueError('No plant loop supplied for water cooling coil')
    plant_loop.addDemandBranchForComponent(clg_coil)

    # set coil cop
    cop = 3.4 if cop is None else cop
    clg_coil.setRatedCoolingCoefficientofPerformance(cop)

    # curve sets
    if type == 'OS default':
        pass  # use OS default curves
    else:  # default curve set
        if model.version() < openstudio.VersionString('3.2.0'):
            clg_coil.setTotalCoolingCapacityCoefficient1(-4.30266987344639)
            clg_coil.setTotalCoolingCapacityCoefficient2(7.18536990534372)
            clg_coil.setTotalCoolingCapacityCoefficient3(-2.23946714486189)
            clg_coil.setTotalCoolingCapacityCoefficient4(0.139995928440879)
            clg_coil.setTotalCoolingCapacityCoefficient5(0.102660179888915)
            clg_coil.setSensibleCoolingCapacityCoefficient1(6.0019444814887)
            clg_coil.setSensibleCoolingCapacityCoefficient2(22.6300677244073)
            clg_coil.setSensibleCoolingCapacityCoefficient3(-26.7960783730934)
            clg_coil.setSensibleCoolingCapacityCoefficient4(-1.72374720346819)
            clg_coil.setSensibleCoolingCapacityCoefficient5(0.490644802367817)
            clg_coil.setSensibleCoolingCapacityCoefficient6(0.0693119353468141)
            clg_coil.setCoolingPowerConsumptionCoefficient1(-5.67775976415698)
            clg_coil.setCoolingPowerConsumptionCoefficient2(0.438988156976704)
            clg_coil.setCoolingPowerConsumptionCoefficient3(5.845277342193)
            clg_coil.setCoolingPowerConsumptionCoefficient4(0.141605667000125)
            clg_coil.setCoolingPowerConsumptionCoefficient5(-0.168727936032429)
        else:
            tccc_name = 'Water to Air Heat Pump Total Cooling Capacity Curve'
            if model.getCurveByName(tccc_name).is_initialized():
                total_cooling_capacity_curve = model.getCurveByName(tccc_name).get()
                total_cooling_capacity_curve = \
                    total_cooling_capacity_curve.to_CurveQuadLinear().get()
            else:
                total_cooling_capacity_curve = openstudio_model.CurveQuadLinear(model)
                total_cooling_capacity_curve.setName(tccc_name)
                total_cooling_capacity_curve.setCoefficient1Constant(-4.30266987344639)
                total_cooling_capacity_curve.setCoefficient2w(7.18536990534372)
                total_cooling_capacity_curve.setCoefficient3x(-2.23946714486189)
                total_cooling_capacity_curve.setCoefficient4y(0.139995928440879)
                total_cooling_capacity_curve.setCoefficient5z(0.102660179888915)
                total_cooling_capacity_curve.setMinimumValueofw(-100)
                total_cooling_capacity_curve.setMaximumValueofw(100)
                total_cooling_capacity_curve.setMinimumValueofx(-100)
                total_cooling_capacity_curve.setMaximumValueofx(100)
                total_cooling_capacity_curve.setMinimumValueofy(0)
                total_cooling_capacity_curve.setMaximumValueofy(100)
                total_cooling_capacity_curve.setMinimumValueofz(0)
                total_cooling_capacity_curve.setMaximumValueofz(100)
            clg_coil.setTotalCoolingCapacityCurve(total_cooling_capacity_curve)

            sccc = 'Water to Air Heat Pump Sensible Cooling Capacity Curve'
            if model.getCurveByName(sccc).is_initialized():
                sensible_cooling_capacity_curve = model.getCurveByName(sccc).get()
                sensible_cooling_capacity_curve = \
                    sensible_cooling_capacity_curve.to_CurveQuintLinear().get()
            else:
                sensible_cooling_capacity_curve = openstudio_model.CurveQuintLinear(model)
                sensible_cooling_capacity_curve.setName(sccc)
                sensible_cooling_capacity_curve.setCoefficient1Constant(6.0019444814887)
                sensible_cooling_capacity_curve.setCoefficient2v(22.6300677244073)
                sensible_cooling_capacity_curve.setCoefficient3w(-26.7960783730934)
                sensible_cooling_capacity_curve.setCoefficient4x(-1.72374720346819)
                sensible_cooling_capacity_curve.setCoefficient5y(0.490644802367817)
                sensible_cooling_capacity_curve.setCoefficient6z(0.0693119353468141)
                sensible_cooling_capacity_curve.setMinimumValueofw(-100)
                sensible_cooling_capacity_curve.setMaximumValueofw(100)
                sensible_cooling_capacity_curve.setMinimumValueofx(-100)
                sensible_cooling_capacity_curve.setMaximumValueofx(100)
                sensible_cooling_capacity_curve.setMinimumValueofy(0)
                sensible_cooling_capacity_curve.setMaximumValueofy(100)
                sensible_cooling_capacity_curve.setMinimumValueofz(0)
                sensible_cooling_capacity_curve.setMaximumValueofz(100)
            clg_coil.setSensibleCoolingCapacityCurve(sensible_cooling_capacity_curve)

            cpcc = 'Water to Air Heat Pump Cooling Power Consumption Curve'
            if model.getCurveByName(cpcc).is_initialized():
                cooling_power_consumption_curve = model.getCurveByName(cpcc).get()
                cooling_power_consumption_curve = \
                    cooling_power_consumption_curve.to_CurveQuadLinear().get()
            else:
                cooling_power_consumption_curve = openstudio_model.CurveQuadLinear(model)
                cooling_power_consumption_curve.setName(cpcc)
                cooling_power_consumption_curve.setCoefficient1Constant(-5.67775976415698)
                cooling_power_consumption_curve.setCoefficient2w(0.438988156976704)
                cooling_power_consumption_curve.setCoefficient3x(5.845277342193)
                cooling_power_consumption_curve.setCoefficient4y(0.141605667000125)
                cooling_power_consumption_curve.setCoefficient5z(-0.168727936032429)
                cooling_power_consumption_curve.setMinimumValueofw(-100)
                cooling_power_consumption_curve.setMaximumValueofw(100)
                cooling_power_consumption_curve.setMinimumValueofx(-100)
                cooling_power_consumption_curve.setMaximumValueofx(100)
                cooling_power_consumption_curve.setMinimumValueofy(0)
                cooling_power_consumption_curve.setMaximumValueofy(100)
                cooling_power_consumption_curve.setMinimumValueofz(0)
                cooling_power_consumption_curve.setMaximumValueofz(100)
            clg_coil.setCoolingPowerConsumptionCurve(cooling_power_consumption_curve)

        # part load fraction correlation curve added as a required curve in OS v3.7.0
        if model.version() > openstudio.VersionString('3.6.1'):
            plfcc_name = 'Water to Air Heat Pump Part Load Fraction Correlation Curve'
            if model.getCurveByName(plfcc_name).is_initialized():
                part_load_correlation_curve = model.getCurveByName(plfcc_name).get()
                part_load_correlation_curve = \
                    part_load_correlation_curve.to_CurveLinear().get()
            else:
                part_load_correlation_curve = openstudio_model.CurveLinear(model)
                part_load_correlation_curve.setName(plfcc_name)
                part_load_correlation_curve.setCoefficient1Constant(0.833746458696111)
                part_load_correlation_curve.setCoefficient2x(0.166253541303889)
                part_load_correlation_curve.setMinimumValueofx(0)
                part_load_correlation_curve.setMaximumValueofx(1)
                part_load_correlation_curve.setMinimumCurveOutput(0)
                part_load_correlation_curve.setMaximumCurveOutput(1)
            clg_coil.setPartLoadFractionCorrelationCurve(part_load_correlation_curve)

    return clg_coil



# coding=utf-8
"""Modules taken from OpenStudio-standards.

https://github.com/NREL/openstudio-standards/blob/master/
lib/openstudio-standards/prototypes/common/objects/Prototype.CoilHeatingElectric.rb
"""
from __future__ import division

from honeybee_openstudio.openstudio import openstudio, openstudio_model

from .utilities import convert_curve_biquadratic, create_curve_biquadratic, \
    create_curve_quadratic
from .schedule import model_add_schedule


def create_coil_heating_electric(
        model, air_loop_node=None, name='Electric Htg Coil',
        schedule=None, nominal_capacity=None, efficiency=1.0):
    """Prototype CoilHeatingElectric object.

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        air_loop_node: [<OpenStudio::Model::Node>] the coil will be placed on
            this node of the air loop.
        name: [String] the name of the system, or None in which case it will be defaulted.
        schedule: [String] name of the availability schedule, or
            [<OpenStudio::Model::Schedule>] Schedule object, or None in which
            case default to always on.
        nominal_capacity: [Double] rated nominal capacity.
        efficiency: [Double] rated heating efficiency.
    """
    htg_coil = openstudio_model.CoilHeatingElectric(model)

    # add to air loop if specified
    if air_loop_node is not None:
        htg_coil.addToNode(air_loop_node)

    # set coil name
    htg_coil.setName(name)

    # set coil availability schedule
    if schedule is None:  # default always on
        coil_availability_schedule = model.alwaysOnDiscreteSchedule()
    elif isinstance(schedule, str):
        if schedule == 'alwaysOffDiscreteSchedule':
            coil_availability_schedule = model.alwaysOffDiscreteSchedule()
        elif schedule == 'alwaysOnDiscreteSchedule':
            coil_availability_schedule = model.alwaysOnDiscreteSchedule()
        else:
            coil_availability_schedule = model_add_schedule(model, schedule)
    else:  # assume that it is an actual schedule object
        coil_availability_schedule = schedule
    htg_coil.setAvailabilitySchedule(coil_availability_schedule)

    # set capacity
    if nominal_capacity is not None:
        htg_coil.setNominalCapacity(nominal_capacity)

    # set efficiency
    if efficiency is not None:
        htg_coil.setEfficiency(efficiency)

    return htg_coil


def create_coil_heating_gas(
        model, air_loop_node=None, name='Gas Htg Coil', schedule=None,
        nominal_capacity=None, efficiency=0.80):
    """Prototype CoilHeatingGas object.

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        air_loop_node: [<OpenStudio::Model::Node>] the coil will be placed on
            this node of the air loop.
        name: [String] the name of the system, or None in which case it will be defaulted.
        schedule: [String] name of the availability schedule, or
            [<OpenStudio::Model::Schedule>] Schedule object, or None in which
            case default to always on.
        nominal_capacity: [Double] rated nominal capacity.
        efficiency: [Double] rated heating efficiency.
    """
    htg_coil = openstudio_model.CoilHeatingGas(model)

    # add to air loop if specified
    if air_loop_node is not None:
        htg_coil.addToNode(air_loop_node)

    # set coil name
    htg_coil.setName(name)

    # set coil availability schedule
    if schedule is None:  # default always on
        coil_availability_schedule = model.alwaysOnDiscreteSchedule()
    elif isinstance(schedule, str):
        if schedule == 'alwaysOffDiscreteSchedule':
            coil_availability_schedule = model.alwaysOffDiscreteSchedule()
        elif schedule == 'alwaysOnDiscreteSchedule':
            coil_availability_schedule = model.alwaysOnDiscreteSchedule()
        else:
            coil_availability_schedule = model_add_schedule(model, schedule)
    else:  # assume that it is an actual schedule object
        coil_availability_schedule = schedule
    htg_coil.setAvailabilitySchedule(coil_availability_schedule)

    # set capacity
    if nominal_capacity is not None:
        htg_coil.setNominalCapacity(nominal_capacity)

    # set efficiency
    if efficiency is not None:
        htg_coil.setGasBurnerEfficiency(efficiency)

    # defaults
    if model.version() < openstudio.VersionString('3.7.0'):
        htg_coil.setParasiticElectricLoad(0.0)
        htg_coil.setParasiticGasLoad(0.0)
    else:
        htg_coil.setOnCycleParasiticElectricLoad(0.0)
        htg_coil.setOffCycleParasiticGasLoad(0.0)

    return htg_coil


def create_coil_heating_water(
        model, hot_water_loop, air_loop_node=None, name='Htg Coil', schedule=None,
        rated_inlet_water_temperature=None, rated_outlet_water_temperature=None,
        rated_inlet_air_temperature=16.6, rated_outlet_air_temperature=32.2,
        controller_convergence_tolerance=0.1):
    """Prototype CoilHeatingWater object.

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        hot_water_loop: [<OpenStudio::Model::PlantLoop>] the coil will be placed
            on the demand side of this plant loop.
        air_loop_node: [<OpenStudio::Model::Node>] the coil will be placed on
            this node of the air loop.
        name: [String] the name of the coil, or nil in which case it will be defaulted.
        schedule: [String] name of the availability schedule, or
            [<OpenStudio::Model::Schedule>] Schedule object, or nil in which case
            default to always on.
        rated_inlet_water_temperature: [Double] rated inlet water temperature in
            degrees Celsius, default is hot water loop design exit temperature.
        rated_outlet_water_temperature: [Double] rated outlet water temperature
            in degrees Celsius, default is hot water loop design return temperature.
        rated_inlet_air_temperature: [Double] rated inlet air temperature in
            degrees Celsius, default is 16.6 (62F).
        rated_outlet_air_temperature: [Double] rated outlet air temperature in
            degrees Celsius, default is 32.2 (90F).
        controller_convergence_tolerance: [Double] controller convergence tolerance.
    """
    htg_coil = openstudio_model.CoilHeatingWater(model)

    # add to hot water loop
    hot_water_loop.addDemandBranchForComponent(htg_coil)

    # add to air loop if specified
    if air_loop_node is not None:
        htg_coil.addToNode(air_loop_node)

    # set coil name
    name = 'Htg Coil' if name is None else name
    htg_coil.setName(name)

    # set coil availability schedule
    if schedule is None:  # default always on
        coil_availability_schedule = model.alwaysOnDiscreteSchedule()
    elif isinstance(schedule, str):
        if schedule == 'alwaysOffDiscreteSchedule':
            coil_availability_schedule = model.alwaysOffDiscreteSchedule()
        elif schedule == 'alwaysOnDiscreteSchedule':
            coil_availability_schedule = model.alwaysOnDiscreteSchedule()
        else:
            coil_availability_schedule = model_add_schedule(model, schedule)
    else:  # assume that it is an actual schedule object
        coil_availability_schedule = schedule
    htg_coil.setAvailabilitySchedule(coil_availability_schedule)

    # rated water temperatures, use hot water loop temperatures if defined
    if rated_inlet_water_temperature is None:
        rated_inlet_water_temperature = \
            hot_water_loop.sizingPlant().designLoopExitTemperature()
        htg_coil.setRatedInletWaterTemperature(rated_inlet_water_temperature)
    else:
        htg_coil.setRatedInletWaterTemperature(rated_inlet_water_temperature)
    if rated_outlet_water_temperature is None:
        rated_outlet_water_temperature = \
            rated_inlet_water_temperature - \
            hot_water_loop.sizingPlant().loopDesignTemperatureDifference()
        htg_coil.setRatedOutletWaterTemperature(rated_outlet_water_temperature)
    else:
        htg_coil.setRatedOutletWaterTemperature(rated_outlet_water_temperature)

    # rated air temperatures
    if rated_inlet_air_temperature is None:
        htg_coil.setRatedInletAirTemperature(16.6)
    else:
        htg_coil.setRatedInletAirTemperature(rated_inlet_air_temperature)
    if rated_outlet_air_temperature is None:
        htg_coil.setRatedOutletAirTemperature(32.2)
    else:
        htg_coil.setRatedOutletAirTemperature(rated_outlet_air_temperature)

    # coil controller properties
    # note These inputs will get overwritten if addToNode or addDemandBranchForComponent
    # is called on the htg_coil object after this
    htg_coil_controller = htg_coil.controllerWaterCoil().get()
    htg_coil_controller.setName('{} Controller'.format(htg_coil.nameString()))
    htg_coil_controller.setMinimumActuatedFlow(0.0)
    if controller_convergence_tolerance is not None:
        htg_coil_controller.setControllerConvergenceTolerance(
            controller_convergence_tolerance)

    return htg_coil


def create_coil_heating_dx_single_speed(
        model, air_loop_node=None, name='1spd DX Htg Coil', schedule=None, type=None,
        cop=3.3, defrost_strategy='ReverseCycle'):
    """Prototype CoilHeatingDXSingleSpeed object.

    Enters in default curves for coil by type of coil

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        air_loop_node: [<OpenStudio::Model::Node>] the coil will be placed on
            this node of the air loop.
        name: [String] the name of the system, or nil in which case it will be defaulted.
        schedule: [String] name of the availability schedule, or
            [<OpenStudio::Model::Schedule>] Schedule object, or nil in which case
            default to always on.
        type: [String] the type of single speed DX coil to reference the correct curve set.
        cop: [Double] rated heating coefficient of performance.
        defrost_strategy: [String] type of defrost strategy. Options are
            reverse-cycle or resistive.
    """
    htg_coil = openstudio_model.CoilHeatingDXSingleSpeed(model)

    # add to air loop if specified
    if air_loop_node is not None:
        htg_coil.addToNode(air_loop_node)

    # set coil name
    htg_coil.setName(name)

    # set coil availability schedule
    if schedule is None:  # default always on
        coil_availability_schedule = model.alwaysOnDiscreteSchedule()
    elif isinstance(schedule, str):
        if schedule == 'alwaysOffDiscreteSchedule':
            coil_availability_schedule = model.alwaysOffDiscreteSchedule()
        elif schedule == 'alwaysOnDiscreteSchedule':
            coil_availability_schedule = model.alwaysOnDiscreteSchedule()
        else:
            coil_availability_schedule = model_add_schedule(model, schedule)
    else:  # assume that it is an actual schedule object
        coil_availability_schedule = schedule
    htg_coil.setAvailabilitySchedule(coil_availability_schedule)

    # set coil cop
    cop = 3.3 if cop is None else cop
    htg_coil.setRatedCOP(cop)

    htg_cap_f_of_temp = None
    htg_cap_f_of_flow = None
    htg_energy_input_ratio_f_of_temp = None
    htg_energy_input_ratio_f_of_flow = None
    htg_part_load_fraction = None
    def_eir_f_of_temp = None

    # curve sets
    if type == 'OS default':
        pass  # use OS defaults
    elif type == 'Residential Central Air Source HP':
        # Performance curves
        # These coefficients are in IP UNITS
        heat_cap_ft_coeffs_ip = [0.566333415, -0.000744164, -0.0000103,
                                 0.009414634, 0.0000506, -0.00000675]
        heat_eir_ft_coeffs_ip = [0.718398423, 0.003498178, 0.000142202,
                                 -0.005724331, 0.00014085, -0.000215321]
        heat_cap_fflow_coeffs = [0.694045465, 0.474207981, -0.168253446]
        heat_eir_fflow_coeffs = [2.185418751, -1.942827919, 0.757409168]
        heat_plf_fplr_coeffs = [0.8, 0.2, 0]
        defrost_eir_coeffs = [0.1528, 0, 0, 0, 0, 0]

        # Convert coefficients from IP to SI
        heat_cap_ft_coeffs_si = convert_curve_biquadratic(heat_cap_ft_coeffs_ip)
        heat_eir_ft_coeffs_si = convert_curve_biquadratic(heat_eir_ft_coeffs_ip)

        htg_cap_f_of_temp = create_curve_biquadratic(
            model, heat_cap_ft_coeffs_si, 'Heat-Cap-fT', 0, 100, 0, 100, None, None)
        htg_cap_f_of_flow = create_curve_quadratic(
            model, heat_cap_fflow_coeffs, 'Heat-Cap-fFF', 0, 2, 0, 2, is_dimensionless=True)
        htg_energy_input_ratio_f_of_temp = create_curve_biquadratic(
            model, heat_eir_ft_coeffs_si, 'Heat-EIR-fT', 0, 100, 0, 100, None, None)
        htg_energy_input_ratio_f_of_flow = create_curve_quadratic(
            model, heat_eir_fflow_coeffs, 'Heat-EIR-fFF', 0, 2, 0, 2, is_dimensionless=True)
        htg_part_load_fraction = create_curve_quadratic(
            model, heat_plf_fplr_coeffs, 'Heat-PLF-fPLR', 0, 1, 0, 1, is_dimensionless=True)

        # Heating defrost curve for reverse cycle
        def_eir_f_of_temp = create_curve_biquadratic(
            model, defrost_eir_coeffs, 'DefrostEIR', -100, 100, -100, 100, None, None)
    elif type == 'Residential Minisplit HP':
        # Performance curves
        # These coefficients are in SI UNITS
        heat_cap_ft_coeffs_si = [1.14715889038462, -0.010386676170938,
                                 0, 0.00865384615384615, 0, 0]
        heat_eir_ft_coeffs_si = [0.9999941697687026, 0.004684593830254383,
                                 5.901286675833333e-05, -0.0028624467783091973,
                                 1.3041120194135802e-05, -0.00016172918478765433]
        heat_cap_fflow_coeffs = [1, 0, 0]
        heat_eir_fflow_coeffs = [1, 0, 0]
        heat_plf_fplr_coeffs = [0.89, 0.11, 0]
        defrost_eir_coeffs = [0.1528, 0, 0, 0, 0, 0]

        htg_cap_f_of_temp = create_curve_biquadratic(
            model, heat_cap_ft_coeffs_si, 'Heat-Cap-fT', -100, 100, -100, 100, None, None)
        htg_cap_f_of_flow = create_curve_quadratic(
            model, heat_cap_fflow_coeffs, 'Heat-Cap-fFF', 0, 2, 0, 2, is_dimensionless=True)
        htg_energy_input_ratio_f_of_temp = create_curve_biquadratic(
            model, heat_eir_ft_coeffs_si, 'Heat-EIR-fT', -100, 100, -100, 100, None, None)
        htg_energy_input_ratio_f_of_flow = create_curve_quadratic(
            model, heat_eir_fflow_coeffs, 'Heat-EIR-fFF', 0, 2, 0, 2, is_dimensionless=True)
        htg_part_load_fraction = create_curve_quadratic(
            model, heat_plf_fplr_coeffs, 'Heat-PLF-fPLR', 0, 1, 0.6, 1, is_dimensionless=True)

        # Heating defrost curve for reverse cycle
        def_eir_f_of_temp = create_curve_biquadratic(
            model, defrost_eir_coeffs, 'Defrost EIR', -100, 100, -100, 100, None, None)
    else:  # default curve set
        coil_name = htg_coil.nameString()
        htg_cap_f_of_temp = openstudio_model.CurveCubic(model)
        htg_cap_f_of_temp.setName('{} Htg Cap Func of Temp Curve'.format(coil_name))
        htg_cap_f_of_temp.setCoefficient1Constant(0.758746)
        htg_cap_f_of_temp.setCoefficient2x(0.027626)
        htg_cap_f_of_temp.setCoefficient3xPOW2(0.000148716)
        htg_cap_f_of_temp.setCoefficient4xPOW3(0.0000034992)
        htg_cap_f_of_temp.setMinimumValueofx(-20.0)
        htg_cap_f_of_temp.setMaximumValueofx(20.0)

        htg_cap_f_of_flow = openstudio_model.CurveCubic(model)
        htg_cap_f_of_flow.setName('{} Htg Cap Func of Flow Frac Curve'.format(coil_name))
        htg_cap_f_of_flow.setCoefficient1Constant(0.84)
        htg_cap_f_of_flow.setCoefficient2x(0.16)
        htg_cap_f_of_flow.setCoefficient3xPOW2(0.0)
        htg_cap_f_of_flow.setCoefficient4xPOW3(0.0)
        htg_cap_f_of_flow.setMinimumValueofx(0.5)
        htg_cap_f_of_flow.setMaximumValueofx(1.5)

        htg_energy_input_ratio_f_of_temp = openstudio_model.CurveCubic(model)
        htg_energy_input_ratio_f_of_temp.setName(
            '{} EIR Func of Temp Curve'.format(coil_name))
        htg_energy_input_ratio_f_of_temp.setCoefficient1Constant(1.19248)
        htg_energy_input_ratio_f_of_temp.setCoefficient2x(-0.0300438)
        htg_energy_input_ratio_f_of_temp.setCoefficient3xPOW2(0.00103745)
        htg_energy_input_ratio_f_of_temp.setCoefficient4xPOW3(-0.000023328)
        htg_energy_input_ratio_f_of_temp.setMinimumValueofx(-20.0)
        htg_energy_input_ratio_f_of_temp.setMaximumValueofx(20.0)

        htg_energy_input_ratio_f_of_flow = openstudio_model.CurveQuadratic(model)
        htg_energy_input_ratio_f_of_flow.setName(
            '{} EIR Func of Flow Frac Curve'.format(coil_name))
        htg_energy_input_ratio_f_of_flow.setCoefficient1Constant(1.3824)
        htg_energy_input_ratio_f_of_flow.setCoefficient2x(-0.4336)
        htg_energy_input_ratio_f_of_flow.setCoefficient3xPOW2(0.0512)
        htg_energy_input_ratio_f_of_flow.setMinimumValueofx(0.0)
        htg_energy_input_ratio_f_of_flow.setMaximumValueofx(1.0)

        htg_part_load_fraction = openstudio_model.CurveQuadratic(model)
        htg_part_load_fraction.setName('{} PLR Correlation Curve'.format(coil_name))
        htg_part_load_fraction.setCoefficient1Constant(0.85)
        htg_part_load_fraction.setCoefficient2x(0.15)
        htg_part_load_fraction.setCoefficient3xPOW2(0.0)
        htg_part_load_fraction.setMinimumValueofx(0.0)
        htg_part_load_fraction.setMaximumValueofx(1.0)

        if defrost_strategy != 'Resistive':
            def_eir_f_of_temp = openstudio_model.CurveBiquadratic(model)
            def_eir_f_of_temp.setName(
                '{} Defrost EIR Func of Temp Curve'.format(coil_name))
            def_eir_f_of_temp.setCoefficient1Constant(0.297145)
            def_eir_f_of_temp.setCoefficient2x(0.0430933)
            def_eir_f_of_temp.setCoefficient3xPOW2(-0.000748766)
            def_eir_f_of_temp.setCoefficient4y(0.00597727)
            def_eir_f_of_temp.setCoefficient5yPOW2(0.000482112)
            def_eir_f_of_temp.setCoefficient6xTIMESY(-0.000956448)
            def_eir_f_of_temp.setMinimumValueofx(-23.33333)
            def_eir_f_of_temp.setMaximumValueofx(29.44444)
            def_eir_f_of_temp.setMinimumValueofy(-23.33333)
            def_eir_f_of_temp.setMaximumValueofy(29.44444)

    if type == 'PSZ-AC':
        htg_coil.setMinimumOutdoorDryBulbTemperatureforCompressorOperation(-12.2)
        htg_coil.setMaximumOutdoorDryBulbTemperatureforDefrostOperation(1.67)
        htg_coil.setCrankcaseHeaterCapacity(50.0)
        htg_coil.setMaximumOutdoorDryBulbTemperatureforCrankcaseHeaterOperation(4.4)
        htg_coil.setDefrostControl('OnDemand')

        def_eir_f_of_temp = openstudio_model.CurveBiquadratic(model)
        def_eir_f_of_temp.setName('{} Defrost EIR Func of Temp Curve'.format(coil_name))
        def_eir_f_of_temp.setCoefficient1Constant(0.297145)
        def_eir_f_of_temp.setCoefficient2x(0.0430933)
        def_eir_f_of_temp.setCoefficient3xPOW2(-0.000748766)
        def_eir_f_of_temp.setCoefficient4y(0.00597727)
        def_eir_f_of_temp.setCoefficient5yPOW2(0.000482112)
        def_eir_f_of_temp.setCoefficient6xTIMESY(-0.000956448)
        def_eir_f_of_temp.setMinimumValueofx(-23.33333)
        def_eir_f_of_temp.setMaximumValueofx(29.44444)
        def_eir_f_of_temp.setMinimumValueofy(-23.33333)
        def_eir_f_of_temp.setMaximumValueofy(29.44444)

    if htg_cap_f_of_temp is not None:
        htg_coil.setTotalHeatingCapacityFunctionofTemperatureCurve(htg_cap_f_of_temp)
    if htg_cap_f_of_flow is not None:
        htg_coil.setTotalHeatingCapacityFunctionofFlowFractionCurve(htg_cap_f_of_flow)
    if htg_energy_input_ratio_f_of_temp is not None:
        htg_coil.setEnergyInputRatioFunctionofTemperatureCurve(
            htg_energy_input_ratio_f_of_temp)
    if htg_energy_input_ratio_f_of_flow is not None:
        htg_coil.setEnergyInputRatioFunctionofFlowFractionCurve(
            htg_energy_input_ratio_f_of_flow)
    if htg_part_load_fraction is not None:
        htg_coil.setPartLoadFractionCorrelationCurve(htg_part_load_fraction)
    if def_eir_f_of_temp is not None:
        htg_coil.setDefrostEnergyInputRatioFunctionofTemperatureCurve(def_eir_f_of_temp)
    htg_coil.setDefrostStrategy(defrost_strategy)
    htg_coil.setDefrostControl('OnDemand')

    return htg_coil


def create_coil_heating_water_to_air_heat_pump_equation_fit(
        model, plant_loop, air_loop_node=None,
        name='Water-to-Air HP Htg Coil', type=None, cop=4.2):
    """Prototype CoilHeatingWaterToAirHeatPumpEquationFit object.

    Enters in default curves for coil by type of coil.

    Args:
        model: [OpenStudio::Model::Model] OpenStudio model object.
        plant_loop: [<OpenStudio::Model::PlantLoop>] the coil will be placed on
            the demand side of this plant loop.
        air_loop_node: [<OpenStudio::Model::Node>] the coil will be placed on
            this node of the air loop.
        name: [String] the name of the system, or None in which case it will
            be defaulted.
        type: [String] the type of coil to reference the correct curve set.
        cop: [Double] rated heating coefficient of performance.
    """
    htg_coil = openstudio_model.CoilHeatingWaterToAirHeatPumpEquationFit(model)

    # add to air loop if specified
    if air_loop_node is not None:
        htg_coil.addToNode(air_loop_node)

    # set coil name
    htg_coil.setName(name)

    # add to plant loop
    if plant_loop is None:
        raise ValueError('No plant loop supplied for heating coil')
    plant_loop.addDemandBranchForComponent(htg_coil)

    # set coil cop
    cop = 4.2 if cop is None else cop
    htg_coil.setRatedHeatingCoefficientofPerformance(cop)

    # curve sets
    if type == 'OS default':
        pass  # use OS default curves
    else:  # default curve set
        if model.version() < openstudio.VersionString('3.2.0'):
            htg_coil.setHeatingCapacityCoefficient1(0.237847462869254)
            htg_coil.setHeatingCapacityCoefficient2(-3.35823796081626)
            htg_coil.setHeatingCapacityCoefficient3(3.80640467406376)
            htg_coil.setHeatingCapacityCoefficient4(0.179200417311554)
            htg_coil.setHeatingCapacityCoefficient5(0.12860719846082)
            htg_coil.setHeatingPowerConsumptionCoefficient1(-3.79175529243238)
            htg_coil.setHeatingPowerConsumptionCoefficient2(3.38799239505527)
            htg_coil.setHeatingPowerConsumptionCoefficient3(1.5022612076303)
            htg_coil.setHeatingPowerConsumptionCoefficient4(-0.177653510577989)
            htg_coil.setHeatingPowerConsumptionCoefficient5(-0.103079864171839)
        else:
            hcc_name = 'Water to Air Heat Pump Heating Capacity Curve'
            hcc = model.getCurveByName(hcc_name)
            if hcc.is_initialized():
                heating_capacity_curve = hcc.get()
                heating_capacity_curve = heating_capacity_curve.to_CurveQuadLinear().get()
            else:
                heating_capacity_curve = openstudio_model.CurveQuadLinear(model)
                heating_capacity_curve.setName(hcc_name)
                heating_capacity_curve.setCoefficient1Constant(0.237847462869254)
                heating_capacity_curve.setCoefficient2w(-3.35823796081626)
                heating_capacity_curve.setCoefficient3x(3.80640467406376)
                heating_capacity_curve.setCoefficient4y(0.179200417311554)
                heating_capacity_curve.setCoefficient5z(0.12860719846082)
                heating_capacity_curve.setMinimumValueofw(-100)
                heating_capacity_curve.setMaximumValueofw(100)
                heating_capacity_curve.setMinimumValueofx(-100)
                heating_capacity_curve.setMaximumValueofx(100)
                heating_capacity_curve.setMinimumValueofy(0)
                heating_capacity_curve.setMaximumValueofy(100)
                heating_capacity_curve.setMinimumValueofz(0)
                heating_capacity_curve.setMaximumValueofz(100)
            htg_coil.setHeatingCapacityCurve(heating_capacity_curve)

            pcc_name = 'Water to Air Heat Pump Heating Power Consumption Curve'
            pcc = model.getCurveByName(pcc_name)
            if pcc.is_initialized():
                heating_power_consumption_curve = pcc.get()
                heating_power_consumption_curve = \
                    heating_power_consumption_curve.to_CurveQuadLinear().get()
            else:
                heating_power_consumption_curve = openstudio_model.CurveQuadLinear(model)
                heating_power_consumption_curve.setName(pcc_name)
                heating_power_consumption_curve.setCoefficient1Constant(-3.79175529243238)
                heating_power_consumption_curve.setCoefficient2w(3.38799239505527)
                heating_power_consumption_curve.setCoefficient3x(1.5022612076303)
                heating_power_consumption_curve.setCoefficient4y(-0.177653510577989)
                heating_power_consumption_curve.setCoefficient5z(-0.103079864171839)
                heating_power_consumption_curve.setMinimumValueofw(-100)
                heating_power_consumption_curve.setMaximumValueofw(100)
                heating_power_consumption_curve.setMinimumValueofx(-100)
                heating_power_consumption_curve.setMaximumValueofx(100)
                heating_power_consumption_curve.setMinimumValueofy(0)
                heating_power_consumption_curve.setMaximumValueofy(100)
                heating_power_consumption_curve.setMinimumValueofz(0)
                heating_power_consumption_curve.setMaximumValueofz(100)
            htg_coil.setHeatingPowerConsumptionCurve(heating_power_consumption_curve)

        # part load fraction correlation curve added as a required curve in OS v3.7.0
        if model.version() > openstudio.VersionString('3.6.1'):
            plfcc_name = 'Water to Air Heat Pump Part Load Fraction Correlation Curve'
            plfcc = model.getCurveByName(plfcc_name)
            if plfcc.is_initialized():
                part_load_correlation_curve = plfcc.get()
                part_load_correlation_curve = \
                    part_load_correlation_curve.to_CurveLinear().get()
            else:
                part_load_correlation_curve = openstudio_model.CurveLinear(model)
                part_load_correlation_curve.setName(plfcc_name)
                part_load_correlation_curve.setCoefficient1Constant(0.833746458696111)
                part_load_correlation_curve.setCoefficient2x(0.166253541303889)
                part_load_correlation_curve.setMinimumValueofx(0)
                part_load_correlation_curve.setMaximumValueofx(1)
                part_load_correlation_curve.setMinimumCurveOutput(0)
                part_load_correlation_curve.setMaximumCurveOutput(1)
            htg_coil.setPartLoadFractionCorrelationCurve(part_load_correlation_curve)

    return htg_coil



# coding=utf-8
"""Methods to read OpenStudio Models into Honeybee Models."""
from __future__ import division
import sys
import os

from ladybug_geometry.geometry3d import Point3D, Face3D
from honeybee.typing import clean_string, clean_ep_string
from honeybee.altnumber import autocalculate
from honeybee.boundarycondition import Outdoors, Surface, boundary_conditions
from honeybee.facetype import face_types
from honeybee.shade import Shade
from honeybee.door import Door
from honeybee.aperture import Aperture
from honeybee.face import Face
from honeybee.room import Room
from honeybee.model import Model
from honeybee_energy.boundarycondition import Adiabatic, OtherSideTemperature
from honeybee_energy.construction.window import WindowConstruction
from honeybee_energy.construction.windowshade import WindowConstructionShade
from honeybee_energy.construction.dynamic import WindowConstructionDynamic

from honeybee_openstudio.openstudio import openstudio, os_vector_len, os_path
from honeybee_openstudio.schedule import extract_all_schedules
from honeybee_openstudio.construction import extract_all_constructions, \
    shade_construction_from_openstudio
from honeybee_openstudio.constructionset import construction_set_from_openstudio
from honeybee_openstudio.load import people_from_openstudio, lighting_from_openstudio, \
    electric_equipment_from_openstudio, gas_equipment_from_openstudio, \
    process_from_openstudio, hot_water_from_openstudio, infiltration_from_openstudio, \
    ventilation_from_openstudio, setpoint_from_openstudio_thermostat, \
    setpoint_from_openstudio_humidistat, daylight_from_openstudio
from honeybee_openstudio.programtype import program_type_from_openstudio
from honeybee_openstudio.hvac.idealair import ideal_air_system_from_openstudio

NATIVE_EP_TOL = 0.01  # native tolerance of E+ in meters
GLASS_CONSTR = (WindowConstruction, WindowConstructionShade, WindowConstructionDynamic)


def face_3d_from_openstudio(os_vertices):
    """Convert an OpenStudio Point3dVector into a Face3D.

    Args:
        os_vertices: An OpenStudio Point3dVector that came from any OpenStudio
            PlanarSurface class.

    Returns:
        A ladybug-geometry Face3D object created from the OpenStudio Point3dVector.
    """
    # create all of the Point3D objects
    vertices = []
    for v in os_vertices:
        vertices.append(Point3D(v.x(), v.y(), v.z()))
    face_3d = Face3D(vertices)

    # sense if the vertices loop back on themselves to cut holes
    separate_holes = False
    for i, pt in enumerate(vertices):
        if i + 2 >= len(vertices):
            break
        for j in range(i + 2, len(vertices)):
            if pt.is_equivalent(vertices[j], NATIVE_EP_TOL):
                separate_holes = True
                break

    # separate the boundary and holes if necessary
    if separate_holes:
        try:
            return face_3d.separate_boundary_and_holes(NATIVE_EP_TOL)
        except AssertionError:  # invalid face to be removed later
            pass
    return face_3d


def shades_from_openstudio(os_shade_group, constructions=None, schedules=None):
    """Convert an OpenStudio ShadingSurfaceGroup into a list of Honeybee Shades.

    Args:
        os_shade_group: An OpenStudio ShadingSurfaceGroup to be converted into
            a list of Honeybee Shades.
        constructions: An optional dictionary of Honeybee Construction objects
            which will be used to assign constructions to the shades.
        schedules: An optional dictionary of Honeybee Schedule objects which will
            be used to assign transmittance schedules to the shades.

    Returns:
        A list of Honeybee Shade objects.
    """
    # get variables that apply to the whole group
    shades = []
    os_site_transform = os_shade_group.siteTransformation()
    shade_type = os_shade_group.shadingSurfaceType()
    is_detached = False if shade_type == 'Space' else True

    # loop through the geometries and create the shade objects
    for os_shade in os_shade_group.shadingSurfaces():
        # create the shade object
        os_vertices = os_site_transform * os_shade.vertices()
        geo = face_3d_from_openstudio(os_vertices)
        shade = Shade(clean_string(os_shade.nameString()), geo, is_detached)
        if os_shade.displayName().is_initialized():
            shade.display_name = os_shade.displayName().get()

        # assign the construction if it exists
        if constructions is not None and not os_shade.isConstructionDefaulted():
            construction = os_shade.construction()
            if construction.is_initialized():
                const_name = \
                    '{} Shade'.format(clean_ep_string(construction.get().nameString()))
                try:
                    shade.properties.energy.construction = constructions[const_name]
                except KeyError:  # make a new shade construction
                    const_obj = construction.get()
                    const = const_obj.to_LayeredConstruction().get()
                    construction = shade_construction_from_openstudio(const)
                    shade.properties.energy.construction = construction
                    constructions[const_name] = const

        # assign the transmittance schedule if it exists
        trans_sch = os_shade.transmittanceSchedule()
        if schedules is not None and trans_sch.is_initialized():
            t_sch_name = clean_ep_string(trans_sch.get().nameString())
            if t_sch_name in schedules:
                try:
                    t_sched = schedules[t_sch_name]
                    shade.properties.energy.transmittance_schedule = t_sched
                except KeyError:
                    pass  # schedule was of a type that could not be loaded

        shades.append(shade)
    return shades


def _extract_sub_surface_bc(os_sub_surface):
    """Get a honeybee boundary condition from an OpenStudio SubSurface."""
    bc = None
    adjacent_sub_surface = os_sub_surface.adjacentSubSurface()
    if adjacent_sub_surface.is_initialized():
        adjacent_sub_surface = adjacent_sub_surface.get()
        adj_door_id = clean_string(adjacent_sub_surface.nameString())
        adj_face_id = None
        if adjacent_sub_surface.surface().is_initialized():
            adj_face_id = clean_string(adjacent_sub_surface.surface().get().nameString())
        adj_room_id = None
        if adjacent_sub_surface.space().is_initialized():
            adj_room_id = clean_string(adjacent_sub_surface.space().get().nameString())
        if adj_face_id is not None and adj_room_id is not None:
            bc = Surface((adj_door_id, adj_face_id, adj_room_id), sub_face=True)
    if bc is None:  # set it to outdoors
        if os_sub_surface.surface().is_initialized():
            os_surface = os_sub_surface.surface().get()
            sun_exposure = True if os_surface.sunExposure() == 'SunExposed' else False
            wind_exposure = True if os_surface.windExposure() == 'WindExposed' else False
            view_factor = os_sub_surface.viewFactortoGround()
            view_factor = view_factor.get() \
                if view_factor.is_initialized() else autocalculate
            bc = Outdoors(sun_exposure, wind_exposure, view_factor)
        else:
            bc = boundary_conditions.outdoors
    return bc


def door_from_openstudio(os_sub_surface, os_site_transform=None, constructions=None):
    """Convert an OpenStudio SubSurface into a Honeybee Door.

    Args:
        os_sub_surface: An OpenStudio SubSurface to be converted into a Honeybee Door.
        site_transform: An optional OpenStudio Transformation object that
            describes how the coordinates of the SubSurface object relate
            to the world coordinate system.
        constructions: An optional dictionary of Honeybee Construction objects
            which will be used to assign a construction to the door.

    Returns:
        A honeybee Door object.
    """
    # create the door object
    os_vertices = os_sub_surface.vertices()
    if os_site_transform is not None:
        os_vertices = os_site_transform * os_vertices
    geo = face_3d_from_openstudio(os_vertices)
    door = Door(clean_string(os_sub_surface.nameString()), geo)
    # assign the display name and type
    if os_sub_surface.displayName().is_initialized():
        door.display_name = os_sub_surface.displayName().get()
    if os_sub_surface.subSurfaceType() == 'GlassDoor':
        door.is_glass = True
    # assign the boundary condition
    door.boundary_condition = _extract_sub_surface_bc(os_sub_surface)
    # assign the construction if it exists
    if constructions is not None and not os_sub_surface.isConstructionDefaulted():
        construction = os_sub_surface.construction()
        if construction.is_initialized():
            const_name = clean_ep_string(construction.get().nameString())
            if const_name in constructions:
                con = constructions[const_name]
                if isinstance(con, GLASS_CONSTR):
                    door.is_glass = True
                door.properties.energy.construction = con
    return door


def aperture_from_openstudio(os_sub_surface, os_site_transform=None, constructions=None):
    """Convert an OpenStudio SubSurface into a Honeybee Aperture.

    Args:
        os_sub_surface: An OpenStudio SubSurface to be converted into
            a Honeybee Aperture.
        site_transform: An optional OpenStudio Transformation object that
            describes how the coordinates of the SubSurface object relate
            to the world coordinate system.
        constructions: An optional dictionary of Honeybee Construction objects
            which will be used to assign a construction to the aperture.

    Returns:
        A honeybee Aperture object.
    """
    # create the door object
    os_vertices = os_sub_surface.vertices()
    if os_site_transform is not None:
        os_vertices = os_site_transform * os_vertices
    geo = face_3d_from_openstudio(os_vertices)
    aperture = Aperture(clean_string(os_sub_surface.nameString()), geo)
    # assign the display name and type
    if os_sub_surface.displayName().is_initialized():
        aperture.display_name = os_sub_surface.displayName().get()
    if os_sub_surface.subSurfaceType() == 'OperableWindow':
        aperture.is_operable = True
    # assign the boundary condition
    aperture.boundary_condition = _extract_sub_surface_bc(os_sub_surface)
    # assign the construction if it exists
    if constructions is not None and not os_sub_surface.isConstructionDefaulted():
        construction = os_sub_surface.construction()
        if construction.is_initialized():
            const_name = clean_ep_string(construction.get().nameString())
            if const_name in constructions:
                con = constructions[const_name]
                if isinstance(con, GLASS_CONSTR):
                    aperture.properties.energy.construction = con
    return aperture


def face_from_openstudio(os_surface, os_site_transform=None, constructions=None):
    """Convert an OpenStudio Surface into a Honeybee Face.

    Args:
        os_surface: An OpenStudio Surface to be converted into a Honeybee Aperture.
        site_transform: An optional OpenStudio Transformation object that
            describes how the coordinates of the SubSurface object relate
            to the world coordinate system.
        constructions: An optional dictionary of Honeybee Construction objects
            which will be used to assign a construction to the aperture.

    Returns:
        A honeybee Aperture object.
    """
    # create the door object
    os_vertices = os_surface.vertices()
    if os_site_transform is not None:
        os_vertices = os_site_transform * os_vertices
    geo = face_3d_from_openstudio(os_vertices)
    face = Face(clean_string(os_surface.nameString()), geo)

    # assign the display name and type
    if os_surface.displayName().is_initialized():
        face.display_name = os_surface.displayName().get()
    face_type = os_surface.surfaceType()
    if os_surface.isAirWall():
        face.type = face_types.air_boundary
    elif 'Wall' in face_type:
        face.type = face_types.wall
    elif 'Floor' in face_type:
        face.type = face_types.floor
    elif 'Roof' in face_type or 'Ceiling' in face_type:
        face.type = face_types.roof_ceiling

    # assign the boundary condition
    bc = None
    surface_bc = os_surface.outsideBoundaryCondition()
    adjacent_surface = os_surface.adjacentSurface()
    if adjacent_surface.is_initialized():
        adjacent_surface = adjacent_surface.get()
        adj_face_id = clean_string(adjacent_surface.nameString())
        adj_room_id = None
        if adjacent_surface.space().is_initialized():
            adj_room_id = clean_string(adjacent_surface.space().get().nameString())
            bc = Surface((adj_face_id, adj_room_id), sub_face=False)
    elif os_surface.isGroundSurface():
        bc = boundary_conditions.ground
    elif surface_bc == 'Adiabatic':
        bc = Adiabatic()
    elif surface_bc == 'OtherSideCoefficients':
        temperature, htc = autocalculate, 0
        if os_surface.surfacePropertyOtherSideCoefficients().is_initialized():
            srf_prop = os_surface.surfacePropertyOtherSideCoefficients().get()
            if not srf_prop.isConstantTemperatureDefaulted():
                temperature = srf_prop.constantTemperature()
            if srf_prop.combinedConvectiveRadiativeFilmCoefficient().is_initialized():
                htc = srf_prop.combinedConvectiveRadiativeFilmCoefficient().get()
        bc = OtherSideTemperature(temperature, htc)
    if bc is None:  # set it to outdoors
        sun_exposure = True if os_surface.sunExposure() == 'SunExposed' else False
        wind_exposure = True if os_surface.windExposure() == 'WindExposed' else False
        view_factor = os_surface.viewFactortoGround()
        view_factor = view_factor.get() \
            if view_factor.is_initialized() else autocalculate
        bc = Outdoors(sun_exposure, wind_exposure, view_factor)

    # assign the construction if it exists
    if constructions is not None and not os_surface.isConstructionDefaulted():
        construction = os_surface.construction()
        if construction.is_initialized():
            const_name = clean_ep_string(construction.get().nameString())
            if const_name in constructions:
                con = constructions[const_name]
                if isinstance(con, GLASS_CONSTR):
                    face.properties.energy.construction = con

    # loop through the sub faces and convert them to Apertures and Doors
    for os_sub_surface in os_surface.subSurfaces():
        sub_surface_type = os_sub_surface.subSurfaceType()
        if 'Door' in sub_surface_type:
            door = door_from_openstudio(
                os_sub_surface, os_site_transform, constructions)
            face.add_door(door)
        else:
            ap = aperture_from_openstudio(
                os_sub_surface, os_site_transform, constructions)
            face.add_aperture(ap)
    return face


def room_from_openstudio(os_space, constructions=None, schedules=None):
    """Convert an OpenStudio Space into a Honeybee Room.

    Args:
        os_space: An OpenStudio Space to be converted into a Honeybee Room.
        constructions: An optional dictionary of Honeybee Construction objects
            which will be used to assign a construction to the room.
        schedules: An optional dictionary of Honeybee Schedule objects which will
            be used to assign schedules to the rooms.

    Returns:
        A honeybee Room object.
    """
    # translate the geometry and the room object
    os_site_transform = os_space.siteTransformation()
    faces = []
    os_surfaces = os_space.surfaces if sys.version_info < (3, 0) else os_space.surfaces()
    for os_surface in os_surfaces:
        face = face_from_openstudio(os_surface, os_site_transform, constructions)
        faces.append(face)
    room_id = clean_string(os_space.nameString())
    if room_id.endswith('_Space'):
        room_id = room_id[:-6]
    room = Room(room_id, faces, tolerance=0.01, angle_tolerance=1.0)

    # assign the display name, multiplier, story, and zone
    if os_space.displayName().is_initialized():
        room.display_name = os_space.displayName().get()
    if os_space.multiplier() != 1:
        room.multiplier = os_space.multiplier()
    inc_flr = os_space.partofTotalFloorArea if sys.version_info < (3, 0) \
        else os_space.partofTotalFloorArea()
    if not inc_flr:
        room.exclude_floor_area = True
    if os_space.buildingStory().is_initialized():
        room.story = os_space.buildingStory().get().nameString()
    if os_space.thermalZone().is_initialized():
        room.zone = os_space.thermalZone().get().nameString()

    # load any shades and assign them to the room
    shades = []
    for os_shade_group in os_space.shadingSurfaceGroups():
        shades.extend(shades_from_openstudio(os_shade_group, constructions, schedules))
    room.add_outdoor_shades(shades)

    # apply all of the loads
    if schedules is not None:
        # assign people
        for os_people in os_space.people():
            people_def = os_people.peopleDefinition()
            if people_def.peopleperSpaceFloorArea().is_initialized():
                room.properties.energy.people = \
                    people_from_openstudio(os_people, schedules)
        # assign lighting
        for os_lights in os_space.lights():
            light_def = os_lights.lightsDefinition()
            if light_def.wattsperSpaceFloorArea().is_initialized():
                room.properties.energy.lighting = \
                    lighting_from_openstudio(os_lights, schedules)
        # assign electric equipment
        for os_equip in os_space.electricEquipment():
            electric_eq_def = os_equip.electricEquipmentDefinition()
            if electric_eq_def.wattsperSpaceFloorArea().is_initialized():
                room.properties.energy.electric_equipment = \
                    electric_equipment_from_openstudio(os_equip, schedules)
        # assign gas equipment
        for os_equip in os_space.gasEquipment():
            electric_eq_def = os_equip.gasEquipmentDefinition()
            if electric_eq_def.wattsperSpaceFloorArea().is_initialized():
                room.properties.energy.gas_equipment = \
                    gas_equipment_from_openstudio(os_equip, schedules)
        # assign hot water
        if sys.version_info >= (3, 0):
            room.properties.energy.service_hot_water = hot_water_from_openstudio(
                os_space.waterUseEquipment(), room.floor_area, schedules)
        # assign process loads
        process_loads = []
        for os_other_eq in os_space.otherEquipment():
            other_eq_def = os_other_eq.otherEquipmentDefinition()
            if other_eq_def.designLevel.empty().is_initialized():
                p_load = process_from_openstudio(other_eq_def, schedules)
                process_loads.append(p_load)
        if len(process_loads) != 0:
            room.properties.energy.process_loads = process_loads
        # assign infiltration
        for os_inf in os_space.spaceInfiltrationDesignFlowRates():
            if os_inf.flowperExteriorSurfaceArea().is_initialized():
                room.properties.energy.infiltration = \
                    infiltration_from_openstudio(os_inf, schedules)
        # assign ventilation
        if os_space.designSpecificationOutdoorAir().is_initialized():
            os_vent = os_space.designSpecificationOutdoorAir().get()
            room.properties.energy.ventilation = \
                ventilation_from_openstudio(os_vent, schedules)
        # assign setpoint
        if os_space.thermalZone().is_initialized():
            os_zone = os_space.thermalZone().get()
            if os_zone.thermostatSetpointDualSetpoint().is_initialized():
                os_thermostat = os_zone.thermostatSetpointDualSetpoint().get()
                setpoint = setpoint_from_openstudio_thermostat(os_thermostat, schedules)
                if os_zone.zoneControlHumidistat().is_initialized():
                    os_humidistat = os_zone.zoneControlHumidistat().get()
                    setpoint = setpoint_from_openstudio_humidistat(
                        os_humidistat, setpoint, schedules)
                room.properties.energy.setpoint = setpoint
        # assign daylight
        if os_vector_len(os_space.daylightingControls()) != 0:
            os_daylight = os_space.daylightingControls()[0]
            room.properties.energy.daylighting_control = \
                daylight_from_openstudio(os_daylight)
    return room


def model_from_openstudio(os_model, reset_properties=False):
    """Convert an OpenStudio Model into a Honeybee Model.

    Args:
        os_model: An OpenStudio Model to be converted into a Honeybee Model.
        reset_properties: Boolean to note whether all energy properties should
            be reset to defaults upon import, meaning that only the geometry and
            boundary conditions are imported from the Openstudio Model. This
            can be particularly useful when importing an openStudio Model that
            originated from an IDF or gbXML since these formats don't support
            higher-level objects like SpaceTypes or ConstructionSets. So it is
            often easier to just import the geometry and reassign properties
            rather than working from a model where all properties are assigned
            to individual objects. (Default: False)

    Returns:
        A honeybee Model.
    """
    # load all of the energy properties from the model
    if reset_properties:
        schedules, constructions = None, None
        construction_sets, program_types = None, None
    else:
        schedules = extract_all_schedules(os_model)
        constructions = extract_all_constructions(os_model, schedules)
        construction_sets = {}
        for os_cons_set in os_model.getDefaultConstructionSets():
            if os_cons_set.nameString() != 'Default Generic Construction Set':
                con_set = construction_set_from_openstudio(os_cons_set, constructions)
                construction_sets[con_set.identifier] = con_set
        program_types = {}
        for os_space_type in os_model.getSpaceTypes():
            program = program_type_from_openstudio(os_space_type, schedules)
            program_types[program.identifier] = program

    # load all of the rooms
    rooms, zone_map = [], {}
    for os_space in os_model.getSpaces():
        room = room_from_openstudio(os_space, constructions, schedules)
        if construction_sets is not None and \
                os_space.defaultConstructionSet().is_initialized():
            os_con_set = os_space.defaultConstructionSet().get()
            try:
                room.properties.energy.construction_set = \
                    construction_sets[os_con_set.nameString()]
            except KeyError:
                pass
        if program_types is not None:
            os_space_type = os_space.spaceType if sys.version_info < (3, 0) \
                else os_space.spaceType()
            if os_space_type.is_initialized():
                os_space_type = os_space_type.get()
                try:
                    room.properties.energy.program_type = \
                        program_types[os_space_type.nameString()]
                except KeyError:
                    pass
        if os_space.thermalZone().is_initialized():
            zone_id = os_space.thermalZone().get().nameString()
            try:
                zone_map[zone_id].append(room)
            except KeyError:  # first Room in the zone
                zone_map[zone_id] = [room]
        rooms.append(room)

    # assign ideal air systems to any relevant zones
    if schedules is not None:
        for os_hvac in os_model.getZoneHVACIdealLoadsAirSystems():
            if os_hvac.thermalZone().is_initialized():
                zone_id = os_hvac.thermalZone().get().nameString()
                hvac = ideal_air_system_from_openstudio(os_hvac, schedules)
                for room in zone_map[zone_id]:
                    room.properties.energy.hvac = hvac

    # load all of the shades
    shades = []
    for os_shade_group in os_model.getShadingSurfaceGroups():
        shading_surface_type = os_shade_group.shadingSurfaceType()
        if shading_surface_type == 'Site' or shading_surface_type == 'Building':
            grp_shades = shades_from_openstudio(os_shade_group, constructions, schedules)
            shades.extend(grp_shades)

    # create the model and return it
    os_building = os_model.getBuilding()
    model_name = os_building.nameString()
    model = Model(clean_string(model_name), rooms=rooms, orphaned_shades=shades,
                  units='Meters', tolerance=0.01, angle_tolerance=1.0)
    model.display_name = model_name
    return model


def model_from_osm(osm_str, reset_properties=False, print_warnings=False):
    """Translate an OSM string to a Honeybee Model.

    Args:
        osm_str: Text string for the contents of an OSM to be converted to a
            Honeybee Model.
        reset_properties: Boolean to note whether all energy properties should be
            reset to defaults upon import, meaning that only the geometry and boundary
            conditions are imported from the Openstudio Model. (Default: False).
        print_warnings: Boolean for whether warnings about unimported objects
            should be printed. (Default: False).
    """
    # get the version translator
    if (sys.version_info < (3, 0)):
        ver_translator = openstudio.VersionTranslator()
    else:
        ver_translator = openstudio.osversion.VersionTranslator()
    os_model = ver_translator.loadModelFromString(osm_str)
    # print errors and warnings from the translation process
    if not os_model.is_initialized():
        errors = '\n'.join(str(err.logMessage()) for err in ver_translator.errors())
        raise ValueError('Failed to load model from OSM.\n{}'.format(errors))
    if print_warnings:
        for warn in ver_translator.warnings():
            print(warn.logMessage())
    # translate the OpenStudio Model to Honeybee
    return model_from_openstudio(os_model.get(), reset_properties)


def model_from_osm_file(osm_file, reset_properties=False, print_warnings=False):
    """Translate an OSM file to a Honeybee Model.

    Args:
        osm_file: Text string for the path to the OSM file to be converted to a
            Honeybee Model.
        reset_properties: Boolean to note whether all energy properties should be
            reset to defaults upon import, meaning that only the geometry and boundary
            conditions are imported from the Openstudio Model. (Default: False).
        print_warnings: Boolean for whether warnings about unimported objects
            should be printed. (Default: False).
    """
    # get the version translator
    assert os.path.isfile(osm_file), 'No file was found at: {}.'.format(osm_file)
    if (sys.version_info < (3, 0)):
        ver_translator = openstudio.VersionTranslator()
    else:
        ver_translator = openstudio.osversion.VersionTranslator()
    os_model = ver_translator.loadModel(os_path(osm_file))
    # print errors and warnings from the translation process
    if not os_model.is_initialized():
        errors = '\n'.join(str(err.logMessage()) for err in ver_translator.errors())
        raise ValueError('Failed to load model from OSM.\n{}'.format(errors))
    if print_warnings:
        for warn in ver_translator.warnings():
            print(warn.logMessage())
    # translate the OpenStudio Model to Honeybee
    return model_from_openstudio(os_model.get(), reset_properties)


def model_from_idf_file(idf_file, reset_properties=False, print_warnings=False):
    """Translate an IDF file to a Honeybee Model.

    Args:
        idf_file: Text string for the path to the IDF file to be converted to a
            Honeybee Model.
        reset_properties: Boolean to note whether all energy properties should be
            reset to defaults upon import, meaning that only the geometry and boundary
            conditions are imported from the Openstudio Model. (Default: False).
        print_warnings: Boolean for whether warnings about unimported objects
            should be printed. (Default: False).
    """
    # get the version translator
    assert os.path.isfile(idf_file), 'No file was found at: {}.'.format(idf_file)
    if (sys.version_info < (3, 0)):
        idf_translator = openstudio.EnergyPlusReverseTranslator()
    else:
        idf_translator = openstudio.energyplus.ReverseTranslator()
    os_model = idf_translator.loadModel(os_path(idf_file))
    # print errors and warnings from the translation process
    if not os_model.is_initialized():
        errors = '\n'.join(str(err.logMessage()) for err in idf_translator.errors())
        raise ValueError('Failed to load model from IDF.\n{}'.format(errors))
    if print_warnings:
        for warn in idf_translator.warnings():
            print(warn.logMessage())
    # translate the OpenStudio Model to Honeybee
    return model_from_openstudio(os_model.get(), reset_properties)


def model_from_gbxml_file(gbxml_file, reset_properties=False, print_warnings=False):
    """Translate a gbXML file to a Honeybee Model.

    Args:
        gbxml_file: Text string for the path to the gbXML file to be converted
            to a Honeybee Model.
        reset_properties: Boolean to note whether all energy properties should be
            reset to defaults upon import, meaning that only the geometry and boundary
            conditions are imported from the Openstudio Model. (Default: False).
        print_warnings: Boolean for whether warnings about unimported objects
            should be printed. (Default: False).
    """
    # get the version translator
    assert os.path.isfile(gbxml_file), 'No file was found at: {}.'.format(gbxml_file)
    if (sys.version_info < (3, 0)):
        gbxml_translator = openstudio.GbXMLReverseTranslator()
    else:
        gbxml_translator = openstudio.gbxml.GbXMLReverseTranslator()
    os_model = gbxml_translator.loadModel(os_path(gbxml_file))
    # print errors and warnings from the translation process
    if not os_model.is_initialized():
        errors = '\n'.join(str(err.logMessage()) for err in gbxml_translator.errors())
        raise ValueError('Failed to load model from OSM.\n{}'.format(errors))
    if print_warnings:
        for warn in gbxml_translator.warnings():
            print(warn.logMessage())
    # remove any shade groups that were translated as spaces
    os_model = os_model.get()
    os_spaces = os_model.getSpaces()
    for os_space in os_spaces:
        os_surfaces = os_space.surfaces if sys.version_info < (3, 0) \
            else os_space.surfaces()
        if os_vector_len(os_surfaces) == 0:
            os_space.remove()
    # translate the OpenStudio Model to Honeybee
    return model_from_openstudio(os_model, reset_properties)



# coding=utf-8
"""Module taken from OpenStudio-standards.

https://github.com/NREL/openstudio-standards/blob/master/
lib/openstudio-standards/thermal_zone/thermal_zone.rb
"""
from __future__ import division

import sys


def thermal_zone_get_outdoor_airflow_rate(thermal_zone):
    """Calculates the zone outdoor airflow requirement (Voz).

    The result will be based on the inputs in the DesignSpecification:OutdoorAir
    objects in all spaces in the zone.

    Args:
        thermal_zone: [OpenStudio::Model::ThermalZone] OpenStudio ThermalZone object.

    Returns:
        [Double] the zone outdoor air flow rate in cubic meters per second (m^3/s).
    """
    tot_oa_flow_rate = 0.0
    spaces = thermal_zone.spaces()

    # Variables for merging outdoor air
    sum_oa_for_people = 0.0
    sum_oa_for_floor_area = 0.0
    sum_oa_rate = 0.0
    sum_oa_for_volume = 0.0

    # Find common variables for the new space
    for space in spaces:
        floor_area = space.floorArea() if sys.version_info >= (3, 0) else space.floorArea
        number_of_people = space.numberOfPeople()
        volume = space.volume()

        dsn_oa = space.designSpecificationOutdoorAir()
        if not dsn_oa.is_initialized():
            continue

        dsn_oa = dsn_oa.get()

        # compute outdoor air rates in case we need them
        oa_for_people = number_of_people * dsn_oa.outdoorAirFlowperPerson()
        oa_for_floor_area = floor_area * dsn_oa.outdoorAirFlowperFloorArea()
        oa_rate = dsn_oa.outdoorAirFlowRate()
        oa_for_volume = volume * dsn_oa.outdoorAirFlowAirChangesperHour() / 3600

        # First check if this space uses the Maximum method and other spaces do not
        if dsn_oa.outdoorAirMethod() == 'Maximum':
            sum_oa_rate += max([oa_for_people, oa_for_floor_area, oa_rate, oa_for_volume])
        elif dsn_oa.outdoorAirMethod() == 'Sum':
            sum_oa_for_people += oa_for_people
            sum_oa_for_floor_area += oa_for_floor_area
            sum_oa_rate += oa_rate
            sum_oa_for_volume += oa_for_volume

    tot_oa_flow_rate += sum_oa_for_people
    tot_oa_flow_rate += sum_oa_for_floor_area
    tot_oa_flow_rate += sum_oa_rate
    tot_oa_flow_rate += sum_oa_for_volume

    return tot_oa_flow_rate


def thermal_zone_get_outdoor_airflow_rate_per_area(thermal_zone):
    """Calculates the zone outdoor airflow requirement and divides by the zone area.

    Args:
        thermal_zone: [OpenStudio::Model::ThermalZone] OpenStudio ThermalZone object.

    Returns:
        [Double] the zone outdoor air flow rate in cubic meters per second
        per floor area(m3/s/m2).
    """
    # Find total area of the zone
    sum_floor_area = 0.0
    for space in thermal_zone.spaces():
        floor_area = space.floorArea() if sys.version_info >= (3, 0) else space.floorArea
        sum_floor_area += floor_area
    # Get the OA flow rate
    tot_oa_flow_rate = thermal_zone_get_outdoor_airflow_rate(thermal_zone)
    # Calculate the per-area value
    tot_oa_flow_rate_per_area = tot_oa_flow_rate / sum_floor_area
    return tot_oa_flow_rate_per_area


def thermal_zone_get_occupancy_schedule(model, thermal_zone):
    """Get the occupancy schedule of the zone.

    Args:
        thermal_zone: [OpenStudio::Model::ThermalZone] OpenStudio ThermalZone object.
    """
    # Get all the occupancy schedules in spaces.
    # Check people added via the SpaceType and hard-assigned to the Space itself.
    occupancy_sch = None
    for space in thermal_zone.spaces():
        # From the space type
        space_type = space.spaceType() if sys.version_info >= (3, 0) else space.spaceType
        if space_type.is_initialized():
            for people in space_type.get().people():
                num_ppl_sch = people.numberofPeopleSchedule()
                if not num_ppl_sch.is_initialized():
                    continue
                occupancy_sch = num_ppl_sch.get()
                break

    if num_ppl_sch is None:  # From the space
        for space in thermal_zone.spaces():
            for people in space.people():
                num_ppl_sch = people.numberofPeopleSchedule()
                if not num_ppl_sch.is_initialized():
                    continue
                occupancy_sch = num_ppl_sch.get()
                break

    # if there is no occupancy, use always off
    if occupancy_sch is None:
        return model.alwaysOffDiscreteSchedule()
    return occupancy_sch
